from toml.encoder import TomlEncoder, _dump_str, unicode


def _dump_str_prefer_multiline(v):
    multilines = v.split("\n")
    if len(multilines) > 1:
        v = '"""\n' + v.replace('"""', '\\"""').strip() + '\n"""'
        v = v.replace('"""', "'''")
        return unicode(v)
    else:
        return _dump_str(v)


class MultilinePreferringTomlEncoder(TomlEncoder):
    def __init__(self, _dict=dict, preserve=False):
        super(MultilinePreferringTomlEncoder, self).__init__(_dict=dict, preserve=preserve)
        self.dump_funcs[str] = _dump_str_prefer_multiline



from __future__ import annotations

import os as _os
from abc import ABC, abstractmethod
from typing import Union


class Serializer:
    file_handlers = {}

    @classmethod
    def register_handler(cls, name: Union[str, list[str]]) -> None:
        def wrapper(handler_cls):
            if isinstance(name, str):
                cls.file_handlers[name] = handler_cls
            elif isinstance(name, list):
                for _name in name:
                    cls.file_handlers[_name] = handler_cls
            else:
                raise TypeError(f"name must be str or list[str], not {type(name)}")

            return handler_cls

        return wrapper

    @classmethod
    def get_handler(cls, output_type):
        return cls.file_handlers[output_type]()

    @classmethod
    def auto_handler(cls, path: str) -> FileHandler:
        name = cls.get_handler_name(path)
        return cls.get_handler(name)

    @classmethod
    def get_handler_name(cls, path: str) -> str:
        return _os.path.splitext(path)[1][1:]

    @classmethod
    def serialize_object(cls, data: dict, path: str) -> None:
        handler = cls.auto_handler(path)
        handler.dump(data, path)

    @classmethod
    def deserialize_object(cls, path: str) -> dict:
        handler = cls.auto_handler(path)
        return handler.load(path)


class FileHandler(ABC):
    @abstractmethod
    def load(self, path: str) -> dict:
        raise NotImplementedError

    @abstractmethod
    def dump(self, data: dict, path: str) -> None:
        raise NotImplementedError



from .handler import FileHandler, Serializer


@Serializer.register_handler("json")
class JsonHandler(FileHandler):
    def load(self, path: str) -> dict:
        import json as _json

        with open(path, "r") as f:
            return _json.load(f)

    def dump(self, data: dict, path: str) -> None:
        import json as _json

        with open(path, "w") as f:
            _json.dump(data, f, sort_keys=False, indent=4)


@Serializer.register_handler("yaml")
class YamlHandler(FileHandler):
    def load(self, path: str) -> dict:
        try:
            from ruamel.yaml import YAML

            _yaml = YAML()
            _yaml.indent(mapping=4, sequence=6, offset=3)
            with open(path, "r") as f:
                return _yaml.load(f)
        except ModuleNotFoundError:
            import yaml as _yaml

            with open(path, "r") as f:
                return _yaml.safe_load(f)

    def dump(self, data: dict, path: str) -> None:
        try:
            from ruamel.yaml import YAML

            _yaml = YAML()
            _yaml.indent(mapping=4, sequence=6, offset=3)
        except ModuleNotFoundError:
            import yaml as _yaml

        with open(path, "w") as f:
            _yaml.dump(data, f)


@Serializer.register_handler("pkl")
class PickleHandler(FileHandler):
    def load(self, path: str) -> dict:
        import pickle as _pickle

        with open(path, "w") as f:
            return _pickle.load(f)

    def dump(self, data: dict, path: str) -> None:
        import pickle as _pickle

        with open(path, "w") as f:
            _pickle.dump(data, f)


@Serializer.register_handler("toml")
class TomlHandler(FileHandler):
    def load(self, path: str) -> dict:
        import toml as _toml

        with open(path, "r") as f:
            return _toml.load(f)

    def dump(self, data: dict, path: str) -> None:
        import toml as _toml

        from .helpers import MultilinePreferringTomlEncoder

        with open(path, "w") as f:
            _toml.dump(data, f, MultilinePreferringTomlEncoder())


@Serializer.register_handler(["", "txt"])
class PlanTextHandler(FileHandler):
    def load(self, path: str) -> dict:
        with open(path, "r") as f:
            return {"data": f.readlines()}

    def dump(self, data: dict, path: str) -> None:
        if "data" not in data:
            raise ValueError("keyword 'data' must be in the dictionary.")
        with open(path, "w") as f:
            f.writelines(data["data"])
